/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.examples;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.BonsaiManager.ConfigurationException;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.util.Pair;
import de.unibi.airobots.bonsai.util.slam.gui.MapPanel;
import de.unibi.airobots.btl.data.map.BinarySlamMap;
import de.unibi.airobots.btl.data.navigation.GlobalPlan;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.vision3d.SRPlaceData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

import java.awt.Component;
import java.awt.Container;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.logging.Level;
import javax.swing.*;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;

/**
 * @author fsiepman
 *
 */
public class ActionRemote extends JPanel implements ActionListener {

    /**
     * Private variables
     */
    private static final long serialVersionUID = 1011534977727520070L;
    private static final Logger logger = Logger.getLogger(ActionRemote.class);
    /**
     * BonSAI sensors and actuators. Not thread safe !! Using a sensor for the
     * SlamMap implies that one additional sensor is needed for any other action
     * that is using the according sensor information.
     */
    private NavigationActuator navigationActuator;
    private Sensor<BinarySlamMap> slamSensor;
    private SpeechActuator speechActuator;
    private Sensor<GlobalPlan> planSensor;
    private Sensor<PositionData> positionSensor;
    private Sensor<SRPlaceData> placeSensor;
    /**
     * Global variables for data display.
     */
    private Double currentXCoordinate = 0.0;
    private Double currentYCoordinate = 0.0;
    private Double currentYawAngle = 0.0;
    private Double goalXCoordinate = 0.0;
    private Double goalYCoordinate = 0.0;
    private Double goalYawAngle = 0.0;
    private int chosenAction = 0;
    private boolean goalReached = false;
    private boolean yawUpdated = false;
    private String currentPlaceLabel = "lab";
    private String lastPlaceLabel = "lab";
    private static final long MOVING_TIMEOUT = 15000;
    protected static final double YAW_TOLERANCE = Math.PI * 0.15;
    private JTextArea yaw;
    private MapPanel mapPanel;

    /**
     * Construct the action remote.
     */
    public ActionRemote() throws IllegalStateException, ConfigurationException,
            URISyntaxException {
        
        // Set up the logger
        BasicConfigurator.configure();
        logger.info("Starting up action remote...");
        BonsaiManager.getInstance().configure(
                "/vol/robocup/trunk/etc/bonsai_config/actionremote.xml");
        logger.info("setting up NavigationActuator...");
        navigationActuator = BonsaiManager.getInstance().createActuator("NavigationActuator", NavigationActuator.class);
        logger.info("setting up SlamSensor...");
        slamSensor = BonsaiManager.getInstance().createSensor("SlamSensor", BinarySlamMap.class);
        logger.info("setting up PositionSensor...");
        positionSensor = BonsaiManager.getInstance().createSensor("PositionSensor", PositionData.class);
        logger.info("setting up SpeechActuator...");
        speechActuator = BonsaiManager.getInstance().createActuator("SpeechActuator", SpeechActuator.class);
        logger.info("setting up PlaceSensor...");
        placeSensor = BonsaiManager.getInstance().createSensor("SRPlaceSensor", SRPlaceData.class);
        logger.info("setting up GlobalPlanSensor...");
        planSensor = BonsaiManager.getInstance().createSensor("GlobalPlanSensor", GlobalPlan.class);
        
        this.setUpGUI();
    }

    private void performRobotAction(int action) {
        //TODO: case switch for different actions switchable via gui
        switch (action) {
            case 1:
                logger.info("Performing action 1 (Identify Scene):");
                try {
                    update();
                    goalReached = navigationActuator.isGoalReached().isStopped();
                    long startMoving = System.currentTimeMillis();
                    while (!goalReached && System.currentTimeMillis() - startMoving < MOVING_TIMEOUT) {
                        try {
                            Thread.sleep(500);
                        } catch (InterruptedException ex) {
                            logger.info("Could not sleep...");
                        }
                        goalReached = navigationActuator.isGoalReached().isStopped();
                    }
                    if (goalReached) {
                        navigationActuator.manualStop();
                        speechActuator.say("I am here to check if I can identify the scene.");
                        try {
                            Thread.sleep(500);
                        } catch (InterruptedException ex) {
                            logger.info("Could not sleep...");
                        }
                        update();
                        if (currentPlaceLabel.equals("")) {
                            speechActuator.say("I don't know this scene.");
                            navigationActuator.manualStop();
                        } else if (currentPlaceLabel.equals("unknown")) {
                            speechActuator.say("I don't know this scene. I need more training.");
                            navigationActuator.manualStop();
                        } else {
                            speechActuator.say("I think this is the " + currentPlaceLabel + ".");
                            navigationActuator.manualStop();
                        }
                        lastPlaceLabel = currentPlaceLabel;
                        update();
                        if (!lastPlaceLabel.equals(currentPlaceLabel)) {
                            logger.info("Last label was " + lastPlaceLabel + " current Label is " + currentPlaceLabel + ".");
                            speechActuator.say("Wait ! Now it looks like the " + currentPlaceLabel + ".");
                        }
                    } else {
                        navigationActuator.manualStop();
                        speechActuator.say("That's a long way. Where did you click again ?");
                    }
                } catch (IOException ex) {
                    logger.error("Could not say anything with speechActuator!");
                }
                break;
            case 2:
                logger.info("Performing action 2 (Ask for label):");
                update();
                try {
                    goalReached = navigationActuator.isGoalReached().isStopped();
                    long startMoving = System.currentTimeMillis();
                    while (!goalReached && System.currentTimeMillis() - startMoving < MOVING_TIMEOUT) {
                        try {
                            Thread.sleep(500);
                        } catch (InterruptedException ex) {
                            logger.info("Could not sleep...");
                        }
                        goalReached = navigationActuator.isGoalReached().isStopped();
                    }
                    if (goalReached) {
                        navigationActuator.manualStop();
                        update();
                        speechActuator.say("Is this the " + currentPlaceLabel + " ?");
                    } else {
                        navigationActuator.manualStop();
                        speechActuator.say("I am a little confused. Where shall I go again ?");
                    }
                } catch (IOException ex) {
                    logger.error("Could not say anything with speechActuator!");
                }
                break;
            case 3:
                logger.info("Performing action 3 (Introduce):");
                update();
                try {
                    speechActuator.say("Let's see.");
                    goalReached = navigationActuator.isGoalReached().isStopped();
                    long startMoving = System.currentTimeMillis();
                    while (!goalReached && System.currentTimeMillis() - startMoving < MOVING_TIMEOUT) {
                        try {
                            Thread.sleep(500);
                        } catch (InterruptedException ex) {
                            logger.info("Could not sleep...");
                        }
                        goalReached = navigationActuator.isGoalReached().isStopped();
                    }
                    if (goalReached) {
                        navigationActuator.manualStop();
                        speechActuator.say("Hello, my name is biron - and I am a very charming robot.");
                        speechActuator.say("My hobbies are driving around in the lab - and watching out for new people around here.");
                        speechActuator.say("In my spare time I try to help researchers here in Bielefeld - with their work.");
                    } else {
                        navigationActuator.manualStop();
                        speechActuator.say("I am a little confused. Where shall I go again ?");
                    }
                } catch (IOException ex) {
                    logger.error("Could not say anything with speechActuator!");
                }
                break;
            default:
                logger.info("No action to perform.");
                try {
                    navigationActuator.manualStop();
                    speechActuator.say("Please select an action from the list.");
                } catch (IOException ex) {
                    java.util.logging.Logger.getLogger(ActionRemote.class.getName()).log(Level.SEVERE, null, ex);
                }
                break;
        }

    }

    @Override
    public void actionPerformed(ActionEvent e) {
        // TODO Button actions...
        if (e.getActionCommand().equals("UPDATE")) {
            update();
        } else if (e.getActionCommand().equals("STOP")) {
            try {
                navigationActuator.manualStop();
            } catch (IOException ex) {
                logger.error("Can't stop!");
            }
            logger.info("STOPPING!");
        } else if (e.getActionCommand().equals("Identify Scene")) {
            chosenAction = 1;
            logger.info("chosenAction is \"Identify Scene\".");
        } else if (e.getActionCommand().equals("Ask for label")) {
            chosenAction = 2;
            logger.info("chosenAction is \"Ask for label\".");
        } else if (e.getActionCommand().equals("Introduce")) {
            chosenAction = 3;
            logger.info("chosenAction is \"Introduce\".");
        } else if (e.getActionCommand().equals("Set YAW")) {
            logger.info("Setting YAW via GUI");
            logger.info("YAW from GUI is: " + yaw.getText());
            currentYawAngle = Double.parseDouble(yaw.getText());
            yawUpdated = true;
        }
    }

    /**
     * Update all data from all sensors.
     */
    private void update() {
        //TODO update relevant sensor info here
        logger.info("Update called...");
        try {
            logger.debug("Reading from slamSensor...");
            BinarySlamMap slamData;
            try {
                slamData = slamSensor.readLast();
            } catch (InterruptedException ex) {
                logger.error("Error reading slamData ! Could not update.");
                return;
            }
            currentXCoordinate = slamData.getX(LengthUnit.METER);
            currentYCoordinate = slamData.getY(LengthUnit.METER);
            if (yawUpdated) {
                logger.info("YAW was set via GUI!");
            } else {
                currentYawAngle = slamData.getYaw(AngleUnit.RADIAN);
            }
            logger.debug("...done with slamSensor update.");
            logger.debug("Checking current goal...");
            NavigationGoalData tmpGoal = navigationActuator.getCurrentGoal();
            goalXCoordinate = tmpGoal.getX(LengthUnit.METER);
            goalYCoordinate = tmpGoal.getY(LengthUnit.METER);
            goalYawAngle = tmpGoal.getYaw(AngleUnit.RADIAN);
            logger.debug("...done with goal update.");
            logger.debug("Updating current place label...");
            logger.info("Prior place label was: " + currentPlaceLabel);
            SRPlaceData tmpLabel;
            try {
                tmpLabel = placeSensor.readLast();
            } catch (InterruptedException ex) {
               logger.error("Error reading placeSensor ! Could not finish update.");
               return;
            }
            currentPlaceLabel = tmpLabel.getSrPlaceLable();
            logger.info("Place label set to " + currentPlaceLabel);
            logger.debug("...done with place label.");
        } catch (IOException ioe) {
            logger.error(ioe.getStackTrace());
            logger.info("Could not read from Sensors!");
        }
    }

    /**
     * Useful for adding action buttons to the layout a la basic remote.
     */
    private static void addComponent(Container cont, GridBagLayout gbl,
            Component c, int x, int y, int width, int height, double weightx,
            double weighty) {
        GridBagConstraints gbc = new GridBagConstraints();
        gbc.fill = GridBagConstraints.BOTH;
        gbc.gridx = x;
        gbc.gridy = y;
        gbc.gridwidth = width;
        gbc.gridheight = height;
        gbc.weightx = weightx;
        gbc.weighty = weighty;
        gbl.setConstraints(c, gbc);
        cont.add(c);
    }

    private void setUpGUI() {
        logger.info("Setting up GUI.");
        JFrame actionGUI = new JFrame("ActionRemote");
        actionGUI.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        actionGUI.setLocation(600, 50);

        JButton updateButton = new JButton("UPDATE");
        updateButton.addActionListener(this);

        JButton stopButton = new JButton("STOP");
        stopButton.addActionListener(this);

        ButtonGroup toggleGroup = new ButtonGroup();

        JToggleButton action1Button = new JToggleButton("Identify Scene");
        action1Button.addActionListener(this);
        toggleGroup.add(action1Button);

        JToggleButton action2Button = new JToggleButton("Ask for label");
        action2Button.addActionListener(this);
        toggleGroup.add(action2Button);

        JToggleButton action3Button = new JToggleButton("Introduce");
        action3Button.addActionListener(this);
        toggleGroup.add(action3Button);

        JButton setYawButton = new JButton("Set YAW");
        setYawButton.addActionListener(this);

        yaw = new JTextArea("3.1415", 1, 5);

        // creating a map panel with SlamSensor, NavigationActuator,
        // PositionSensor, GlobalPlanSensor and update time (long)
        mapPanel = new MapPanel(slamSensor, navigationActuator, positionSensor, planSensor, 500);

        // create a world clicked listener for the map panel
        MapPanel.WorldClickListener mapClickedListener = new MapPanel.WorldClickListener() {

            @Override
            public void worldClicked(Pair<Double> worldCoordinates, MouseEvent mouseEvent) {

                update();
                NavigationGoalData goal = new NavigationGoalData();

                goal.setX(worldCoordinates.getFirst(), LengthUnit.METER);
                goal.setY(worldCoordinates.getSecond(), LengthUnit.METER);
                // set YAW straight in front
                goal.setYaw(currentYawAngle, AngleUnit.RADIAN);

                if (mouseEvent.getButton() == MouseEvent.BUTTON2
                        || mouseEvent.getButton() == MouseEvent.BUTTON3) {
                    goal.setType(GoalType.GLOBAL);
                }

                // YAW tolerance should be < Math.PI/2
                goal.setYawTolerance(YAW_TOLERANCE, AngleUnit.RADIAN);

                try {
                    navigationActuator.setGoal(goal);
                } catch (IOException e) {
                    logger.warn("Could not set target!", e);
                }

                performRobotAction(chosenAction);

            }
        };

        // add the world clicked listener
        mapPanel.addWorldClickListener(mapClickedListener);


        // getting container of the gui
        Container guiMainContainer = actionGUI.getContentPane();
        // setting layout for gui
        GridBagLayout gbl = new GridBagLayout();
        guiMainContainer.setLayout(gbl);

        // adding buttons to layout with custom function
        // cont, gbl, comp, x, y, width, height, weightx, weighty
        addComponent(guiMainContainer, gbl, updateButton, 1, 0, 1, 1, 1.0, 0);
        addComponent(guiMainContainer, gbl, stopButton, 1, 1, 1, 1, 1.0, 0);
        addComponent(guiMainContainer, gbl, setYawButton, 1, 3, 1, 1, 1.0, 0);
        addComponent(guiMainContainer, gbl, yaw, 2, 3, 1, 1, 1.0, 0);
        addComponent(guiMainContainer, gbl, action1Button, 2, 0, 1, 1, 1.0, 0);
        addComponent(guiMainContainer, gbl, action2Button, 2, 1, 1, 1, 1.0, 0);
        addComponent(guiMainContainer, gbl, action3Button, 2, 2, 1, 1, 1.0, 0);
        addComponent(guiMainContainer, gbl, mapPanel, 3, 0, 570, 500, 1.0, 0);

        actionGUI.pack();
        actionGUI.setVisible(true);

    }

    /**
     * @param args
     */
    public static void main(String[] args) {
        try {
            // Invoke the remote
            new ActionRemote();
        } catch (IllegalStateException ex) {
            logger.error("Illegal State: " + ex.getStackTrace());
            System.exit(-1);
        } catch (ConfigurationException ex) {
            logger.error(ex.getMessage() + "\n" + ex.getCause().getLocalizedMessage());
            System.exit(-1);
        } catch (URISyntaxException ex) {
            logger.error("URI Error: " + ex.getStackTrace());
            System.exit(-1);
        }

    }
}
