/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills;

import de.unibi.airobots.bonsai.core.Actuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.BonsaiManager.ConfigurationException;
import de.unibi.airobots.bonsai.core.CoreObjectFactory.CoreObjectCreationException;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.configuration.ConfigurationParser;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredActuator;
import de.unibi.airobots.bonsai.core.configuration.ConfiguredSensor;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Set;
import org.apache.log4j.Logger;

/**
 * A class to initialize and manage sensors and actuators.
 * 
 * @author nkoester
 * @author lkettenb
 */
public class SensorActuatorManager {
    /**
     * Hash map containing all actuators.
     */
    private HashMap<String, Actuator> actuatorMap =
            new HashMap<String, Actuator>();
    
    /**
     * Hash map containing all sensors.
     */
    private HashMap<String, Sensor<?>> sensorMap = 
            new HashMap<String, Sensor<?>>();
    
    /**
     * The log.
     */
    private Logger logger = Logger.getLogger(SensorActuatorManager.class);

    
    /**
     * Convenience Constructor.
     */
    public SensorActuatorManager() {
        
    }
    
    
    /**
     * Initialize all sensors and actuators.
     * 
     * @param bonsaiConfigXMLfile Name of the Bonsai configuration XML file.
     * @return <code>true</code> on success, <code>false</code> otherwise.
     */
    public boolean initialize(String bonsaiConfigXMLfile) {
        boolean success;
        try {
            URL configXmlURL = new URL("file:" + bonsaiConfigXMLfile);
            
            BonsaiManager.getInstance().configure(configXmlURL.toURI());
            ConfigurationParser parser = 
                    BonsaiManager.getInstance().getParser();
                    
            
            logger.debug("Bonsai config XML found and configured. - " + 
                    configXmlURL.toString());

            Set<ConfiguredSensor> sensorSet = parser.getConfiguredSensors();
            Set<ConfiguredActuator> actuatorSet =
                    parser.getConfiguredActuators();
            
            success = createSensors(sensorSet);
            success |= createActuators(actuatorSet);
        } catch (MalformedURLException e) {
            logger.error("Bonsai Config URL error: " + e.getMessage());
            success = false;
        } catch (IllegalStateException e) {
            logger.error("Bonsai Config XML error: " + e.getMessage());
            success = false;
        } catch (ConfigurationException e) {
            logger.error("Bonsai Config XML parse error: " + e.getMessage());
            success = false;
        } catch (URISyntaxException e) {
            logger.error("Bonsai Config XML uri-syntax error: "
                    + e.getMessage());
            success = false;
        }
        return success;
    }
    
    
    /**
     * Tries to create all sensors.
     * 
     * @param sensorSet Set of sensors read from BonSAI configuration.
     * @return <code>true</code> on success, <code>false</code> otherwise.
     */
    private boolean createSensors(Set<ConfiguredSensor> sensorSet) {
        boolean success = true;
        for (ConfiguredSensor configuredSensor : sensorSet) {
            // Define sensor key and sensor class
            String sensorKey = configuredSensor.getKey();
            Class<?> sensorClass;
            try {
                sensorClass = Class.forName(
                        configuredSensor.getDataTypeClassName());
            } catch (ClassNotFoundException e) {
                logger.error("Class not found for sensor:" + sensorKey);
                success = false;
                continue;
            }

            // Factory able to create a sensor with the unique key?
            if (!BonsaiManager.getInstance().canCreateSensor(
                    sensorKey, sensorClass)) {
                logger.error("Cannot create sensor: " + sensorKey);
                success = false;
                continue;
            }

            // Try to create sensor
            try {
                sensorMap.put(sensorKey, 
                        BonsaiManager.getInstance().createSensor(
                        sensorKey, sensorClass));
                logger.debug("Loading and adding sensor: " + sensorKey);
            } catch (IllegalArgumentException e) {
                logger.error("Bonsai sensor creation error (IllegalArgument"
                        + "Exception): "
                        + e.getMessage());
                success = false;
            } catch (CoreObjectCreationException e) {
                logger.error("Bonsai sensor creation error (CoreObject"
                        + "CreationException): "
                        + e.getMessage());
                success = false;
            } 
        }
        return success;
    }
    
    
    /**
     * Tries to create all actuators.
     * 
     * @param actuatorSet Set of actuators read from BonSAI configuration.
     * @return <code>true</code> on success, <code>false</code> otherwise.
     */
    private boolean createActuators(Set<ConfiguredActuator> actuatorSet) {
        boolean success = true;
        for (ConfiguredActuator configuredActuator : actuatorSet) {
            // Define actuator key and actuator class
            String actuatorKey = configuredActuator.getKey();
            Class<? extends Actuator> actuatorClass;
            try {
                actuatorClass = (Class<? extends Actuator>) Class.forName(
                        configuredActuator.getActuatorClassName());
            } catch (ClassNotFoundException e) {
                logger.error("Class not found for actuator:" + actuatorKey);
                success = false;
                continue;
            }

            // Factory able to create a actuator with the unique key?
            if (!BonsaiManager.getInstance().canCreateActuator(
                    actuatorKey, actuatorClass)) {
                logger.error("Cannot create actuator: " + actuatorKey);
                success = false;
                continue;
            }

            try {
                actuatorMap.put(
                        actuatorKey, 
                        BonsaiManager.getInstance().createActuator(
                        actuatorKey, actuatorClass));
                logger.debug("Loading and adding actuator: " + actuatorKey);
            } catch (IllegalArgumentException e) {
                logger.error("Bonsai actuator creation error (Illegal"
                        + "ArgumentException): "
                        + e.getMessage());
                success = false;
            } catch (CoreObjectCreationException e) {
                logger.error("Bonsai actuator creation error (Core"
                        + "ObjectCreationException): "
                        + e.getMessage());
                success = false;
            } 
        }
        return success;
    }
    

    /**
     * Returns a reference to an actuator (if present).
     * 
     * @param <T> The type of the actuator.
     * @param actuatorName Name of the actuator.
     * @return The requested actuator or null, if no actuator with the given
     * name exists.
     */
    public <T extends Actuator> T getManagedActuator(String actuatorName) {
        if (actuatorMap.containsKey(actuatorName)) {
            // ToDo: Catch bad cast exception
            return (T) actuatorMap.get(actuatorName);
        } else {
            return null;
        }
    }

    
    /**
     * Returns a reference to a sensor (if present).
     * 
     * @param <T> The type of the actuator.
     * @param sensorName Name of the sensor.
     * @return The requested sensor or null, if no sensor with the given
     * name exists.
     */
    public <T extends Sensor> T getManagedSensor(String sensorName) {
        if (sensorMap.containsKey(sensorName)) {
            // ToDo: Catch bad cast exception
            return (T) sensorMap.get(sensorName);
        } else {
            return null;
        }
    }
    
    
    /**
     * List of all initialized sensors.
     * @return String list of all initialized actuators.
     */
    public String[] getSensors() {
        String[] ret = new String[sensorMap.size()];
        ret = sensorMap.keySet().toArray(ret);
        return ret;
    }
    
    
    /**
     * List of all initialized actuators.
     * @return String list of all initialized actuators.
     */
    public String[] getActuators() {
        String[] ret = new String[actuatorMap.size()];
        ret = actuatorMap.keySet().toArray(ret);
        return ret;
    }
}
