/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.bonsai.strategies.NavigationStrategy;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter.ReachStatus;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import java.io.IOException;
import org.apache.log4j.Logger;

/**
 * Use this state to drive to a given (global) position. The state will look for 
 * the newest {@link NavigationGoalData} in scene memory and will than try to 
 * reach this position.
 * 
 * <pre>
 * Special exit statuses: <code>error</code> if unable to access memeory or 
 * uneable to reach goal.
 * </pre>
 * 
 * @author nkoester
 * @author lkettenb
 */
public class DriveToPosition extends AbstractState {
    
    /**
     * How long to wait for new data?
     */
    private static final int TIMEOUT = 3000;
    
    /**
     * Position data with target information.
     */
    private NavigationGoalData navigationGoalData = null;
    
    /*
     * Sensors used by this state.
     */
    private Sensor<PositionData> positionSensor;
    private Sensor<SpeedData> speedSensor;
    private Sensor<GlobalPlannerStateChange> gplanChangeSensor;
    private Sensor<NavigationGoalData> memoryNavigationGoalDataSensor;
    
    /*
     * Actuators used by this state.
     */
    private NavigationActuator navActuator;
    
    
    /**
     * The logger.
     */
    private Logger logger = getLogger();
    
    /**
     * Constructor.
     * @param aStateMachine The state machine to use.
     */
    public DriveToPosition(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }
    
    
    @Override
    public ExitStatus init() throws InitException {
        // Initialize actuators
        navActuator = getManagedActuator("NavigationActuator");
        
        // Initialize sensors
        positionSensor = getManagedSensor("PositionSensor");
        gplanChangeSensor = getManagedSensor("GlobalPlannerStateChangeSensor");
        speedSensor = getManagedSensor("SpeedSensor");
        memoryNavigationGoalDataSensor = getManagedSensor("BtlMemorySensor");
        
        try {
            navigationGoalData = memoryNavigationGoalDataSensor.readLast(TIMEOUT);
        } catch (IOException ex) {
            logger.warn(ex.getMessage());
        } catch (InterruptedException ex) {
            logger.warn(ex.getMessage());
        }
        
        if (navigationGoalData == null) {
            return ExitStatus.ERROR;
        }
        
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus execute() throws ExecuteException {
        NavigationStrategy navigationStrategy = 
                new NavigationStrategy(navActuator, positionSensor, 
                        speedSensor, gplanChangeSensor);
        
        try {
            ReachStatus reachStatus = 
                    navigationStrategy.driveTo(navigationGoalData);
            if (reachStatus.compareTo(ReachStatus.REACHED) == 0) {
                return ExitStatus.SUCCESS;
            } else {
                logger.error("Error while trying to reach goal.");
                return ExitStatus.ERROR;
            }
        } catch (IOException e) {
            logger.error("Exception while trying to reach goal: " + e.getMessage());
            return ExitStatus.ERROR;
        }
    }

    @Override
    public ExitStatus end() throws EndException {
        return ExitStatus.SUCCESS;
    }
}
