/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.actuators.BtlMemoryActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.FaceIdentificationSensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.bonsai.skills.helper.PersonManager;
import de.unibi.airobots.btl.Type;
import de.unibi.airobots.btl.data.person.PersonData;
import de.unibi.airobots.btl.data.person.PersonDataList;
import java.io.IOException;
import org.apache.log4j.Logger;

/**
 * This state will add a new person to a PersonDataList and learn its face.
 * 
 * If a PersonDataList has been written to SceneMemory before it will be used.
 * 
 * <pre>
 * Special exit statuses: <code>error</code> if no person is standing in front
 * of the robot.
 * </pre>
 * 
 * @see PersonDataList
 * 
 * @author lkettenb
 */
public class StoreNewPerson extends AbstractState {
    /*
     * What the robot will say.
     */
    private static final String STATE_INTRO1 = 
            "Please look into my eye while I tell you a bit about myself.";
    private static final String STATE_INTRO2 = 
            "My name is ToBI and I'm from Bielefeld.";
    private static final String STATE_INTRO3 = 
            "What I'm doing right now is trying to learn your face.";
    private static final String STATE_INTRO4 = 
            "So please give me a few more seconds.";
    private static final String STATE_FINISHED = 
            "Ok. From now on i am able to recognize you!";
    
    /*
     * Actuators used by this state.
     */
    private SpeechActuator speechActuator;
    private BtlMemoryActuator memoryActuator;
    
    /*
     * Sensors used by this state.
     */
    private FaceIdentificationSensor faceSensor;
    private Sensor<PersonList> personSensor;
    private Sensor<PersonDataList> personDataListSensor;
    
    /**
     * The logger.
     */
    private Logger logger = getLogger();
    
    /**
     * Additional time to learn person.
     */
    private static final int TIME_TO_LEARN = 3000;
    
    /**
     * List with person data. Will be read from and saved to the scene memory.
     */
    private PersonDataList personDataList;
    
    /**
     * True, if person data was read from memory.
     */
    private boolean personDataReadFromMem = false;
    
    /**
     * Error message if memory cannot be read.
     */
    private static final String ERROR_MEM = "Unable to access memory. "
            + "PersonDataList will be overwritten. ";
    
    /**
     * Constructor.
     * @param aStateMachine The state machine to use.
     */
    public StoreNewPerson(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }
    
    @Override
    public ExitStatus init() throws InitException {
        // Initialize actuators
        speechActuator = getManagedActuator("SpeechActuator");
        memoryActuator = 
                getManagedActuator("BtlMemoryActuator");
        // Initialize sensors
        faceSensor = (FaceIdentificationSensor)
                getManagedSensor("FaceIdentificationSensor");
        personSensor = getManagedSensor("PersonSensor");
        personDataListSensor = 
                getManagedSensor("BtlMemorySensor");
        
        // look for existing de.unibi.airobots.btl.data.PersonDataList 
        // in scene memory
        if (personDataListSensor.hasNext()) {
            try {
                personDataList = personDataListSensor.readLast();
            } catch (IOException ex) {
                logger.warn(ERROR_MEM + ex.getMessage());
            } catch (InterruptedException ex) {
                logger.warn(ERROR_MEM + ex.getMessage());
            }
            personDataReadFromMem = true;
        } else {
            personDataList = new PersonDataList();
        }
        
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus execute() throws ExecuteException {
        // get the next person
        PersonData newPerson = 
                PersonManager.getNextPersonInFront(personSensor);

        if (newPerson == null) {
            return ExitStatus.ERROR;
        }

        // set id
        newPerson.setId(personDataList.size());

        // start learning
        try {
            faceSensor.activate();
            faceSensor.startLearning(newPerson.getId());
            say(STATE_INTRO1);
            say(STATE_INTRO2);
            say(STATE_INTRO3);
            say(STATE_INTRO4);

            // giving more time to learn the person
            sleep(TIME_TO_LEARN);

            // stop learning
            faceSensor.stopLearning();
            say(STATE_FINISHED);

            // disable face detection
            faceSensor.deactivate();
        } catch (IOException ex) {
            // Face sensor does not work. That is bad!
            throw new ExecuteException(ex.getMessage());
        } catch (InterruptedException ex) {
            // Not so bad. Sleep did not work.
            logger.warn(ex.getMessage());
        }
        // storing learned person
        logger.info("Add person to list: " + newPerson);
        personDataList.addPersonData(newPerson);

        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus end() throws EndException {
        try {
            if (personDataReadFromMem) {
                memoryActuator.replace(personDataList);
                logger.debug("Replaced existing person data.");
            } else { 
                memoryActuator.insert(personDataList, PersonDataList.class);
                logger.debug("Inserted new person data.");
            }
        } catch (IOException ex) {
            // Too bad. Unable to write memory.
            throw new EndException(ex.getMessage());
        } catch (Type.NoSourceDocumentException ex) {
            // Too bad. Unable to write memory.
            throw new EndException(ex.getMessage());
        }
        return ExitStatus.SUCCESS;
    }
    
    
    /**
     * Use speech actuator to say something and catch IO exception.
     * 
     * @param text Text to be said.
     */
    private void say(String text) {
        try {
            speechActuator.say(text);
        } catch (IOException ex) {
            // Not so bad. The robot just says nothing.
            logger.warn(ex.getMessage());
        }
    }
    
}
