/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.core.SensorListener;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import de.unibi.airobots.btl.data.vision1d.LaserData;
import de.unibi.airobots.btl.units.LengthUnit;
import de.unibi.airobots.btl.units.SpeedUnit;

import java.io.IOException;
import org.apache.log4j.Logger;

/**
 * Use this state to wait until the door opens.
 * 
 * @author nkoester
 * @author lkettenb
 */
public class WaitForOpenDoor extends AbstractState implements
		SensorListener<LaserData> {

	// Managers and actuators
	private SpeechActuator speechActuator;
	private Sensor<SpeedData> speedSensor;
	private Sensor<LaserData> laserSensor;

	/*
	 * What the robot will say.
	 */
	private static final String KNOCK_KNOCK = "Knock knock, please open the door.";

	/**
	 * Old scan values (will be compared with new values).
	 */
	private double[] oldValues = null;

	/**
	 * Indicates whether the door is open or not.
	 */
	private boolean isOpen = false;

	/**
	 * We expect that the door is in front of the robot. So we only have a look
	 * at the 10 scan values in front of the robot.
	 * 
	 * SCAN_FROM = 361 / 2 - 5, with 361 = number of scan values.
	 * 
	 * @see de.unibi.airobots.btl.data.LaserData#getScanValues()
	 * @see de.unibi.airobots.bonsai.skills.states.WaitForOpenDoor#
	 *      newDataAvailable(de.unibi.airobots.btl.data.LaserData)
	 */
	private static final int SCAN_FROM = 176;

	/**
	 * We expect that the door is in front of the robot. So we only have a look
	 * at the 10 scan values in front of the robot.
	 * 
	 * SCAN_TO = 361 / 2 + 5, with 361 = number of scan values.
	 * 
	 * @see de.unibi.airobots.btl.data.LaserData#getScanValues()
	 * @see de.unibi.airobots.bonsai.skills.states.WaitForOpenDoor#
	 *      newDataAvailable(de.unibi.airobots.btl.data.LaserData)
	 */
	private static final int SCAN_TO = 186;

	/**
	 * The maximum error of scan values (door open?).
	 */
	private static final double MAX_ERROR = 1.5;

	/**
	 * Indicates how long the thread will sleep if the door is closed.
	 */
	public static final int TIME_TO_SLEEP = 500;

	/**
	 * The logger.
	 */
	private Logger logger = super.getLogger();

	/**
	 * Constructor.
	 * 
	 * @param aStateMachine
	 *            The state machine to use.
	 */
	public WaitForOpenDoor(SkillStateMachine aStateMachine) {
		super(aStateMachine);
	}

	@Override
	public ExitStatus init() throws InitException {
		// Initialize actuators
		speechActuator = getManagedActuator("SpeechActuator");

		// Initialize sensors
		laserSensor = getManagedSensor("LaserSensor");
		speedSensor = getManagedSensor("SpeedSensor");

		laserSensor.addSensorListener(this);

		try {
			speechActuator.say(KNOCK_KNOCK, true);
		} catch (IOException ex) {
			logger.warn(ex.getMessage());
		}

		return ExitStatus.SUCCESS;
	}

	@Override
	public ExitStatus execute() throws ExecuteException {
		if (!isOpen) {
			try {
				sleep(TIME_TO_SLEEP);
			} catch (InterruptedException ex) {
				logger.warn(ex.getMessage());
			}

			setProcessingStatus(getWaitProcessingMessage());
			return ExitStatus.SUCCESS;
		}
		return ExitStatus.SUCCESS;
	}

	@Override
	public ExitStatus end() throws EndException {
		laserSensor.removeSensorListener(this);
		return ExitStatus.SUCCESS;
	}

	@Override
	public void newDataAvailable(LaserData newData) {
		try {
			if (speedSensor.readLast().getSpeedTranslation(
					SpeedUnit.METER_PER_SEC) == 0) {
				double[] scanValues = newData.getScanValues(LengthUnit.METER);

				if (oldValues == null) {
					oldValues = scanValues;
				} else {
					double error = 0;
					for (int i = SCAN_FROM; i < SCAN_TO; i++) {
						error = +oldValues[i] - scanValues[i];
					}

					assert error != 0;

					if (error < 0 && Math.abs(error) > MAX_ERROR) {
						logger.debug("Door open. Error: " + error);
						isOpen = true;
					}

					if (error > 0 && Math.abs(error) > MAX_ERROR) {
						logger.debug("Door closed. Error: " + error);
					}
					oldValues = scanValues;
				}
			}
		} catch (IOException e) {
			logger.error("IOException while reading new data. " + e);
		} catch (InterruptedException e) {
			logger.error("InterruptedException while reading new data. " + e);
		}
	}
}
