/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states.dialog;

import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.bonsai.skills.helper.SpeechManager;
import de.unibi.airobots.bonsai.skills.helper.SpeechManager.NonTerminalListener;
import de.unibi.airobots.btl.data.speechrec.Utterance;
import java.io.IOException;
import org.apache.log4j.Logger;


/**
 * Use this class to wait for a audible interrupt signal (e.g. "stop").
 * 
 * @author lkettenb
 */
public class WaitForInterrupt extends AbstractState 
    implements NonTerminalListener {
    /*
     * What the robot will say.
     */
    private static final String CONFIRM_STOP =  "O.K. I will stop.";
    
    /*
     * Actuators used by this state.
     */
    private SpeechActuator speechActuator;
    
    /*
     * Sensors used by this state.
     */
    private Sensor<Utterance> speechSensor;
    
    /**
     * The logger.
     */
    private Logger logger = super.getLogger();
    
    /**
     * SpeechManager used for understanding etc.
     */
    private SpeechManager speechManager;
    
    /**
     * True if the robot understood "stop".
     */
    private volatile boolean stopCommand = false;
    
    /**
     * Constructor.
     * @param aStateMachine The state machine to use.
     */
    public WaitForInterrupt(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }

    @Override
    public ExitStatus init() throws InitException {
        speechSensor = getManagedSensor("SpeechSensor");
        speechActuator = getManagedActuator("SpeechActuator");
        speechManager = new SpeechManager(speechSensor, speechActuator);
        speechManager.addNonTerminalListener(this, "stop");
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus execute() throws ExecuteException {
        if (stopCommand) {
            logger.debug("Interrupted by user.");
            try {
                speechActuator.say(CONFIRM_STOP);
            } catch (IOException ex) {
                logger.warn(ex.getMessage());
            }
            return ExitStatus.SUCCESS;
        }
        
        setProcessingStatus(getWaitProcessingMessage());
        return ExitStatus.SUCCESS;
    }

    @Override
    public ExitStatus end() throws EndException {
        return ExitStatus.SUCCESS;
    }
    
    @Override
    public void processNonTerminal() {
        stopCommand = true;
    }
}
