/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.view;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

/**
 * GUI for the skill state machine.
 * 
 * @author lkettenb
 */
public class StateMachineViewer extends JFrame implements ActionListener {
    
    /**
     * List with all StateMachineViewerListener.
     * 
     * @see StateMachineViewerListener
     */
    private ArrayList<StateMachineViewerListener> listeners = 
            new ArrayList<StateMachineViewerListener>();
    
    /*
     * Constant for coordinates, sizes, etc.
     */
    private static final int FRAME_X = 50;
    private static final int FRAME_Y = 50;
    private static final int FRAME_WIDTH = 1000;
    private static final int FRAME_HEIGHT = 600;
    private static final String TITLE = "State Machine Viewer";
    
    /*
     * Default paths to files.
     */
    private String defaultConfigFilePath = "/";
    private String defaultTaskFilePath = "/";
    
    
    /*
     * Panles used by the GUI.
     */
    private JPanel topPanel;
    private JPanel centerPanel;
    private JPanel bottomPanel;
    private SensorsAndActuatorsListPanel pSensorsAndActuators = 
            new SensorsAndActuatorsListPanel();
    private ConfigAndTaskLoadPanel pConfigAndTaskLoad;
    private StateMachineControlPanel pStateMachineControl = 
            new StateMachineControlPanel(this);
    private CurrentAndNextStatePanel pCurrentAndNextState = 
            new CurrentAndNextStatePanel();
    
    
    
    
    /**
     * Convenience constructor.
     */
    public StateMachineViewer() {
        super(TITLE);
        setupUI();
    }
    
    
    /**
     * Use this constructor if the path to task and configuration file is
     * already known.
     * 
     * @param pathToConfig Path to configuration file.
     * @param pathToTask Path to task file.
     */
    public StateMachineViewer(String pathToConfig, String pathToTask) {
        super(TITLE);
        defaultConfigFilePath = pathToConfig;
        defaultTaskFilePath = pathToTask;
        
        setupUI();
    }
    
    
    /**
     * Connect new StateMachineViewerListener to this StateMachineViewer.
     * 
     * @see StateMachineViewerListener
     * 
     * @param listener New listener to add.
     */
    public void addListener(StateMachineViewerListener listener) {
        listeners.add(listener);
    }
    
    
    /**
     * Setup the user interface.
     */
    private void setupUI() {
        setBounds(FRAME_X, FRAME_Y, FRAME_WIDTH, FRAME_HEIGHT);
        
        topPanel = setupTop();
        centerPanel = setupCenter();
        bottomPanel = setupBottom();
        
        Container con = getContentPane();
        
        con.add(topPanel, BorderLayout.NORTH);
        con.add(centerPanel, BorderLayout.CENTER);
        con.add(bottomPanel, BorderLayout.SOUTH);
        
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setVisible(true); 
    }
    
    
    /**
     * Setup the top panel with file choosers to load configuration files.
     * 
     * @return The top panel.
     */
    private JPanel setupTop() {
        pConfigAndTaskLoad = new ConfigAndTaskLoadPanel(this, 
                defaultConfigFilePath, defaultTaskFilePath);
        return pConfigAndTaskLoad;
    }
    
    
    /**
     * Setup the center panel with current 
     * state/next states/sensors/actuators overview.
     * 
     * @return The center panel.
     */
    private JPanel setupCenter() {
        JPanel tmpPanel = new JPanel(new FlowLayout());
        tmpPanel.add(pCurrentAndNextState);
        tmpPanel.add(pSensorsAndActuators);
        return tmpPanel;
    }
    
    
    /**
     * Setup the bottom panel with event chooser.
     * 
     * @return The bottom panel.
     */
    private JPanel setupBottom() {
        pStateMachineControl.setEnabled(false);
        return pStateMachineControl;
    }

    
    @Override
    public void actionPerformed(ActionEvent e) {
        if (e.getActionCommand().equals(
                ConfigAndTaskLoadPanel.AC_LOAD_CONFIG_FILE)) {
            for (StateMachineViewerListener listener : listeners) {
                listener.loadConfig(pConfigAndTaskLoad.getPathToConfigFile());
            }
        }
        if (e.getActionCommand().equals(
                ConfigAndTaskLoadPanel.AC_LOAD_TASK_FILE)) {
            for (StateMachineViewerListener listener : listeners) {
                listener.loadTask(pConfigAndTaskLoad.getPathToTaskFile());
            }
        }
        if (e.getActionCommand().equals(StateMachineControlPanel.AC_FIRE_EVENT)) {
            for (StateMachineViewerListener listener : listeners) {
                listener.fireEvent(pStateMachineControl.getCurrentSelectedEvent());
            }
        }
        if (e.getActionCommand().equals(StateMachineControlPanel.AC_START_EVENT)) {
            for (StateMachineViewerListener listener : listeners) {
                listener.executeStateMachine();
            }
        }
    }
    
    
    /**
     * Tell user interface if BonSAI configuration was loaded successfully.
     * 
     * @param success <code>true</code> if BonSAI configuration was 
     * loaded successfully, <code>else</code> otherwise.
     */
    public void setLoadConfigSuccessful(boolean success) {
        if (success) {
            pConfigAndTaskLoad.setTaskFileChooserEnabled(true);
            pConfigAndTaskLoad.setConfigChooserEnabled(false);
        } else {
            JOptionPane.showMessageDialog(this,
                "Error configuring a sensor. See log for details.",
                "BonSAI configuration - XML parse error",
                JOptionPane.ERROR_MESSAGE);
        }
    }
    
    
    /**
     * Tell user interface if task file was loaded successfully.
     * 
     * @param success <code>true</code> if task file was 
     * loaded successfully, <code>else</code> otherwise.
     */
    public void setLoadTaskSuccessful(boolean success) {
        if (success) {
            pStateMachineControl.setEnabled(true);
            pConfigAndTaskLoad.setTaskFileChooserEnabled(false);
        } else {
            JOptionPane.showMessageDialog(this,
                "Error while decoding SCXML file. See log for details.",
                "SCXML parse error",
                JOptionPane.ERROR_MESSAGE);
        }
    }
    
    /**
     * Update the sensors and actuators list.
     * 
     * @param sensors List with all sensors.
     * @param actuators List with all actuators.
     */
    public void updateSensorsActuatorsListModel(String[] sensors, 
            String[] actuators) {
        pSensorsAndActuators.updateSensorsActuatorsList(sensors, actuators);
    }
    
    
    /**
     * Update list of currently possible events.
     * 
     * @param eventList List of currently possible events.
     */
    public void updateEventList(String[] eventList) {
        pStateMachineControl.updateEventList(eventList);
    }
    
    /**
     * Clear list of current active states.
     * 
     * @param state ID of a current active state.
     */
    public void removeCurrentActiveState(String state) {
       pCurrentAndNextState.removeCurrentActiveState(state);
    }
    
    
    /**
     * Add a current active state to the list.
     * 
     * @param state ID of a current active state.
     */
    public void addCurrentActiveState(String state) {
        pCurrentAndNextState.addCurrentActiveState(state);
    }
    
    /**
     * Update the next states list.
     * 
     * @param states List with all states.
     */
    public void updateNextStatesList(String[] states) {
        pCurrentAndNextState.updateNextStatesList(states);
    }
}
