/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf;

import java.lang.reflect.InvocationTargetException;
import java.util.Set;
import java.util.concurrent.CopyOnWriteArraySet;

import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.event.system.SystemEventListener;

import org.apache.log4j.Logger;

import Ice.ConnectFailedException;

/**
 * A decorator for {@link RemoteServer}s that consumes all errors related to
 * connection problems, reports them to a
 * {@link RemoteServerReconnectableManager} instance and waits for a new
 * {@link RemoteServer} from the manager to retry the call until it succeeds.
 * 
 * <p>
 * This class can be used if the remote server manages no state and and the
 * client does not want to know anything about possible connection errors or
 * reconnections to a server. If this is needed, a different error handling
 * approach is needed.
 * </p>
 * TODO:
 * <ul>
 * <li>add max tries variable to avoid endless locking</li>
 * </ul>
 * 
 * @author jwienke
 */
public class ConnectionErrorResistantRemoteServer extends RemoteServerWrapper {

    private Logger logger = Logger.getLogger(getClass());

    /**
     * A mirrored list of {@link SystemEventListener}s that will be reassigned
     * to the remote server instance if the old one died.
     */
    private Set<SystemEventListener> listeners =
            new CopyOnWriteArraySet<SystemEventListener>();

    /**
     * Constructor.
     * 
     * @param manager
     *            {@link RemoteServerReconnectableManager} managing this
     *            instance
     */
    ConnectionErrorResistantRemoteServer(
            RemoteServerReconnectableManager manager) {
        super(null, manager);
    }

    // -------- wrapper interface ----------------------------------------------

    // CHECKSTYLE:OFF we really want to throw everything that is possible here
    // because XCF already does this.
    @Override
    protected Object reflectiveCallMethod(RemoteServerMethod method,
            Object... args) throws RuntimeException, UserException,
            MethodNotFoundException {
        // CHECKSTYLE:ON

        assert method != null;

        int iteration = 0;
        do {

            iteration++;
            logger.debug("Calling method with key " + method + ", iteration "
                    + iteration);

            synchronized (getRemoteServerLock()) {

                logger.debug("Got remoteServerLock");

                try {
                    return invokeWrappedServerMethod(method, args);
                } catch (InvocationTargetException e) {

                    logger.warn("Received exception "
                            + "while calling wrapped remote server.", e);

                    if (e.getCause() instanceof ConnectFailedException) {

                        // if there was an ice based error, this server seems to
                        // be unreachable in the network. Therefore globally
                        // mark this server as disconnected and wait for a new
                        // one.
                        markWrappedServerAsDisconnected();

                        try {
                            waitForWrappedRemoteServer();
                        } catch (InterruptedException e1) {
                            throw new RuntimeException(e1);
                        }

                    } else if (e.getCause() instanceof UserException) {
                        throw (UserException) e.getCause();
                    } else if (e.getCause() instanceof MethodNotFoundException) {
                        throw (MethodNotFoundException) e.getCause();
                    } else {
                        logger.fatal("Unexpected exception in "
                                + "wrapped remote server.", e);
                        assert false;
                        throw new RuntimeException(e);
                    }

                }

            }

            // TODO check maximum number of tries
        } while (true);

    }

    /**
     * Resets all listeners after a remote serve reset.
     */
    @Override
    protected void postProcessServerReset() {
        // TODO this can potentially crash again if this remote server died
        // again
        // catch this exception and set remote server to null again
        // reset listeners
        for (SystemEventListener l : listeners) {
            getWrappedRemoteServer().addListener(l);
        }
    }

    /**
     * Mirrors the list of listeners locally to reassign them after
     * reconnection.
     * 
     * @param l
     *            listener to add
     */
    @Override
    public void addListener(SystemEventListener l) {
        // I don't know how to lock this operation so that it is atomic without
        // a deadlock.
        // Right now a listener is possibly set two times in the wrapped remote
        // server if a reset occurs in the reflective call
        listeners.add(l);
        super.addListener(l);
    }

    /**
     * Mirrors the list of listeners locally to reassign them after
     * reconnection.
     * 
     * @param l
     *            listener to remove
     */
    @Override
    public void removeListener(SystemEventListener l) {
        listeners.remove(l);
        super.removeListener(l);
    }

}
